/*
 * Copyright (c) 2007-2009 Sergey Matveychuk
 *      Yandex, LLC.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the company nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <freeradius-client.h>
#include "dhcprelya.h"

#define RC_CONFIG_FILE "/usr/local/etc/radiusclient/radiusclient.conf"

static rc_handle *rh;
static char **only_for;
static unsigned only_for_num=0;

void *
send_acct(void *packet)
{
    struct dhcp_packet *dhcp = (struct dhcp_packet *)packet;
    char str[100], buf[100];
    VALUE_PAIR *send, **vp;
    u_long nas_port;
    
    /* Ignore all packets w/o an assigned IP address */
    if(dhcp->yiaddr.s_addr == 0)
	return NULL;

    send = NULL;
    vp = &send;
    sprintf(str, "Acct-Status-Type=Start");
    logd(LOG_DEBUG, "%s", str);
    if(rc_avpair_parse(rh, str, vp) < 0) {
	logd(LOG_ERR, "radius_plugin: rc_avpair_parse() error");
	goto end;
    }
    vp = &send->next;
    sprintf(str, "User-Name=%s", print_mac(dhcp->chaddr, buf));
    logd(LOG_DEBUG, "%s", str);
    if(rc_avpair_parse(rh, str, vp) < 0) {
	logd(LOG_ERR, "radius_plugin: rc_avpair_parse() error");
	goto end;
    }
    vp = &send->next;

    sprintf(str, "Framed-Ip-Address=%s", print_ip(dhcp->yiaddr.s_addr, buf));
    logd(LOG_DEBUG,"%s", str);
    if(rc_avpair_parse(rh, str, vp) < 0) {
	logd(LOG_ERR, "radius_plugin: rc_avpair_parse() error");
	goto end;
    }

    nas_port = dhcp->yiaddr.s_addr;
    if(rc_acct(rh, nas_port, send) != OK_RC) {
	logd(LOG_ERR, "radius_plugin: rc_acct() error");
	goto end;
    }
    logd(LOG_DEBUG, "OK");
end:
    free(packet);
    return NULL;
}

int
radius_plugin_init(plugin_options_head_t *options_head)
{
    char *rc_conf = RC_CONFIG_FILE;
    struct plugin_options *opts, *opts_tmp;
    char *p, *p1, buf[16];
    int i, n=0;
    ip_addr_t addr, mask;

    if((rh = rc_read_config(rc_conf)) == NULL) {
	logd(LOG_ERR, "radius_plugin: can't read radiusclient.conf");
	return 0;
    }
    if(rc_read_dictionary(rh, rc_conf_str(rh, "dictionary")) != 0) {
	logd(LOG_ERR, "radius_plugin: error reading radius dictionary");
	return 0;
    }
    SLIST_FOREACH_SAFE(opts, options_head, next, opts_tmp) {
	if((p = strchr(opts->option_line, '=')) == NULL) {
	    logd(LOG_ERR, "radius_plugin: syntax error at line: %s", opts->option_line);
	    return 0;
	}
	*p = '\0'; p++;
	if(strcasecmp(opts->option_line, "only_for") == 0) {
	    only_for_num = 0;
	    for(i=0; i<strlen(p); i++)
		if(p[i] == ' ' || p[i] == '\t')
		    n++;
	    only_for = malloc(sizeof(char*)*(n+1));
	    if(only_for == NULL) {
		logd(LOG_ERR, "radius_plugin: malloc error");
		return 0;
	    }
	    while((p1 = strsep(&p, " \t")) != NULL) {
		only_for[only_for_num] = malloc(strlen(p1)+1);
		if(only_for[only_for_num] == NULL) {
		    logd(LOG_ERR, "radius_plugin: malloc error");
		    return 0;
		}
		strlcpy(only_for[only_for_num], p1, strlen(p1));
		only_for_num++;
	    }
	} else if(strcasecmp(opts->option_line, "bind_interface") == 0) {
	    if(!get_ip(p, &addr, &mask)) {
		logd(LOG_ERR, "radius_plugin: interface %s not found", p);
		return 0;
	    }
	    if(rh->this_host_bind_ipaddr == NULL) {
		rh->this_host_bind_ipaddr = malloc(sizeof(*rh->this_host_bind_ipaddr));
		if(rh->this_host_bind_ipaddr == NULL) {
		    logd(LOG_ERR, "radius_plugin: malloc error");
		    return 0;
		}
	    }
	    if(rh->this_host_bind_ipaddr != NULL)
		*rh->this_host_bind_ipaddr = htonl(addr);
	} else {
	    logd(LOG_ERR, "radius_plugin: unknown option at line: %s", opts->option_line);
	    return 0;
	}
	free(opts->option_line);
	SLIST_REMOVE(options_head, opts, plugin_options, next);
	free(opts);
    }
    for(i=0; i<only_for_num; i++)
	logd(LOG_DEBUG, "only_for: %s", only_for[i]);
    logd(LOG_DEBUG, "radius_plugin: bind address: %s", print_ip(ntohl(rc_own_bind_ipaddress(rh)), buf));
    return 1;
}

int
radius_plugin_send_to_client(const struct sockaddr_in *server,
	const struct interface *intf, char **buf, unsigned *psize)
{
    pthread_t tid;
    int i, found=0;
    char *b;

    /* Look for interfaces we should do radius request */
    for(i=0; i<only_for_num; i++)
	if(strcmp(only_for[i], intf->name) == 0)
	    found=1;

    if(only_for_num == 0 || found) {
	b = malloc(*psize-(ETHER_HDR_LEN+DHCP_UDP_OVERHEAD));
	if(b == NULL) {
	    logd(LOG_ERR, "radius_plugin: malloc error");
	    return 0;
	}
	memcpy(b, *buf+ETHER_HDR_LEN+DHCP_UDP_OVERHEAD, *psize-(ETHER_HDR_LEN+DHCP_UDP_OVERHEAD));
	pthread_create(&tid, NULL, send_acct, b);
	pthread_detach(tid);
    }

    return 1;
}

struct plugin_data radius_plugin = {
	"radius",
	radius_plugin_init,
	NULL,
	NULL,
	NULL,
	radius_plugin_send_to_client
};
